/*

Copyright (c) 2003, Mo DeJong

This file is part of Source-Navigator.

Source-Navigator is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your option)
any later version.

Source-Navigator is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with Source-Navigator; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.



*/

/*
 * phpbrowser.l
 *
 * Copyright (C) 2003 Mo DeJong
 *
 * Description:
 * Lex input file for an php language processor.
 */

%{

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "snptools.h"
#include "lexinput.h"
#include "longstr.h"
#include "srchtbl.h"
#include "tcl.h"

#undef YY_INPUT
#define YY_INPUT(buf,r,ms) (r = sn_encoded_input(buf, ms))

#undef yywrap
#define YY_SKIP_YYWRAP

typedef enum {
    OPEN_PAREN,    /* '(' */
    CLOSE_PAREN,   /* ')' */
    OPEN_BRACE,    /* '{' */
    CLOSE_BRACE,   /* '}' */
    OPEN_BRACKET,  /* '[' */
    CLOSE_BRACKET, /* ']' */
    SEMICOLON,     /* ';' */
    COMMA,         /* ',' */
    PERIOD,        /* '.' */

    VARIABLE,      /* $var */
    ASSIGNMENT_OPERATOR, /* '=' */
    ASSIGNMENT_OPERATORS, /* = -= += and so on */
    COMPARISON_OPERATORS, /* == === != >= and so on */
    INCREMENT_OPERATORS,  /* ++ and -- */
    REFERENCE_OPERATOR,   /* & */

    SOMEWORD,          /* a sequence of text */
    KEYWORD,       /* a key sequence in this language */
    FUNCTION_KEYWORD, /* The keyword "function" */
    GLOBAL_KEYWORD, /* The keyword "global" */
    INCLUDE_KEYWORD, /* The keyword "include" */

    VDOUBLE_QUOTED_STRING, /* "I am a $variable string" */
    DOUBLE_QUOTED_STRING, /* "I am a double quoted string" */
    SINGLE_QUOTED_STRING,  /* 'I am a double quoted string' */

    UNKNOWN,        /* Token that parser does not know about */
    COMMENT,        /* Emit COMMENT token only when debugging */
    HTML            /* Emit HTML token only when debugging */
} TokenType;

typedef enum {
    VAR_READ,
    VAR_WRITE,
    VAR_READWRITE
} VarAccess;

typedef struct Token {
    TokenType type;
    char * strval; /* String value of token, NULL if single character */
    long start_line;
    int start_column;
    long end_line;
    int end_column;
    struct Token* vars; /* for variables embedded in a DOUBLE_QUOTED_STRING token */
    struct Token* next;
} Token;

typedef struct NestedBracket {
    int count;
    int prefix_incr;
    struct Token* var;
    struct NestedBracket* next;
} NestedBracket;

typedef struct NestedFunction {
    char* name;
    char* args;
    int   high_line;
    int   high_col_start;
    int   high_col_end;
    int   line_start;
    int   line_end;
    int   col_start;
    int   col_end;
    int   brace_count;
    struct NestedFunction* next;
} NestedFunction;

/* Uncomment this to see debug output in token parsing stage */
/* #define TOKEN_DEBUG */

#ifdef TOKEN_DEBUG
FILE* tokenout = NULL;
#endif

static char * token_dump_file = NULL;
static int highlight_file = 0;

Token* tokens_head = NULL;
Token* tokens_tail = NULL;
int token_index = 0; /* current offset from original token_head */

static char group[] = "php";

#define MAX_SIZE 512

/* FIXME: Can't pass NULL or "" as current function (core dump) */
static char* current_function = (char *) NULL;
static char* current_function_args = (char *) NULL;

/* line number where highlight starts and ends */
static int  current_function_highlight_line;

/* in "function fog() {}" column of 'f' in "fog" */
static int  current_function_highlight_column_start;

/* in "function fog() {}" column of 'g' in "fog" */
static int  current_function_highlight_column_end;

/* line where "function" appears */
static int  current_function_line_start;

/* line where closing brace of function appears */
static int  current_function_line_end;

/* in "function fog() {}" column of 'f' in "function" */
static int  current_function_column_start;

/* in "function fog() {}" column of '}' in "{}" */
static int  current_function_column_end;

static int  current_function_brace_count;

static NestedFunction* function_nest_head = NULL;
static NestedFunction* function_nest_tail = NULL;


static int current_array_bracket_count = 0;
static Token* current_array_token = NULL;
static int current_array_prefix_incr = 0;
static NestedBracket* array_nest_head = NULL;
static NestedBracket* array_nest_tail = NULL;

static int result;

static SearchTable * global_var_table = (SearchTable *) NULL;
static SearchTable * super_global_var_table;

/* Stores the contents of a special processing mode over
 * multiple lines/rules.
 */
LongString mode_buff;
long mode_start_line;
int mode_start_column;

#define COMMENT_DUMP 0
#define DQSTRING_DUMP 0
#define SQSTRING_DUMP 0
#define HDSTRING_DUMP 0
#define HTML_DUMP 0
#define MATCH_DUMP 0

static YY_BUFFER_STATE original_buffer;

static char* heredoc_id = (char *) NULL; /* End marker for a heredoc string */

static Token* embedded_dq_string_vars_head = NULL;
static Token* embedded_dq_string_vars_tail = NULL;

#if MATCH_DUMP
static void matched_pattern(char * pattern, char * text);
#endif

void enter_html_mode();

static char* modestring();

static void FreeGlobalEntry(SearchEntry *entry);

void FreeToken(Token* tok);
Token* pop_head_token();
void free_head_token();
void push_bracket_counter();
void pop_bracket_counter();
void push_function();
void pop_function();
void append_token(TokenType type,
                  char* strval,
                  long start_line,
                  int start_column,
                  long end_line,
                  int end_column);

void append_dqstring_var_token(char* var,
                           long start_line,
                           int start_column,
                           long end_line,
                           int end_column);

void append_dqstring_token(char* strval,
                           long start_line,
                           int start_column,
                           long end_line,
                           int end_column);

void emit_var_access_for_dqstring(Token* tok);

char * TokenTypeToString(Token *tok);

int get_num_tokens_matched(char *match);

void emit_function_declaration();
void emit_comment();
void emit_dqstring();
void emit_sqstring();
void emit_hdstring();
void emit_html();
void emit_var_access(Token* tok, VarAccess acc);

int yywrap() { return 1; }

%}

%x COMMENT_MODE
%x HTML_MODE
%x DQSTRING
%s SQSTRING
%s HDSTRING
%x PHP
%x TOKEN

ws		[ \t]
wsn		[ \t\n]
symbol		[a-zA-Z0-9_]+

varname		{symbol}
someword	{symbol}
token		{symbol}
heredoc-id	{symbol}

assignment-operators	("+="|"-="|"*="|"/="|".="|"%="|"&="|"|="|"^="|"<<="|">>=")
comparison-operators	("=="|"==="|"!="|"!=="|"<>"|"<"|">"|"<="|">=")
increment-operators     ("++"|"--")

any-tokens	(" "|" "{symbol}" "|" "{symbol}" "({symbol}" ")*)

include-keywords	("include"|"include_once"|"require"|"require_once")

keywords	("and"|"or"|"xor"|"array"|"as"|"break"|"case"|"cfunction"|"class"|"const"|"continue"|"declare"|"default"|"die"|"do"|"echo"|"else"|"elseif"|"empty"|"enddeclare"|"endfor"|"endforeach"|"endif"|"endswitch"|"endwhile"|"eval"|"exit"|"extends"|"for"|"foreach"|"if"|"isset"|"list"|"new"|"old_function"|"print"|"return"|"static"|"switch"|"unset"|"use"|"var"|"while"|"__FUNCTION__"|"__FILE__"|"__CLASS__"|"__LINE__")

dq-string-token  "DOUBLE_QUOTED_STRING"
sq-string-token  "SINGLE_QUOTED_STRING"
literal-token ({someword}|{dq-string-token}|{sq-string-token})
include-string-token ({dq-string-token}|{sq-string-token})

function-argument-token  ("VARIABLE"|"REFERENCE_OPERATOR VARIABLE"|"VARIABLE ASSIGNMENT_OPERATOR "{literal-token})
function-argument-tokens (" "|" "{function-argument-token}" "("COMMA "{function-argument-token}" ")*)

variable-tokens	(" VARIABLE "|" VARIABLE "("COMMA VARIABLE ")+)


var-array-index-token ("SOMEWORD"|"VDOUBLE_QUOTED_STRING"|"DOUBLE_QUOTED_STRING"|"SINGLE_QUOTED_STRING"|"VARIABLE")
var-array-empty-tokens "OPEN_BRACKET CLOSE_BRACKET"
var-array-index-tokens "OPEN_BRACKET "{var-array-index-token}" CLOSE_BRACKET")
var-array-token ({var-array-empty-tokens}|{var-array-index-tokens}

var-array-tokens (" "|" "{var-array-token}" "({var-array-token}" ")*)

%%
    /* Start in HTML mode */
    highlight_file = (int) sn_getopt(SN_OPT_HIGHLIGHT);
    token_dump_file = (char *) sn_getopt(SN_OPT_DUMP_TOKENS);
    enter_html_mode();

<PHP>"?>" { /* PHP -> HTML mode */
    #if MATCH_DUMP
    matched_pattern("?>", yytext);
    #endif

    enter_html_mode();

    if (token_dump_file) {
      mode_buff.append( &mode_buff,
          yytext, yyleng );
    }

    sn_advance_column(2);
}

<HTML_MODE>{
  [^\<\n]* {
    #if MATCH_DUMP
    matched_pattern("[^\\<\\n]*", yytext);
    #endif

    #if HTML_DUMP
    fprintf(stderr, "html(1) \"%s\", %d\n", yytext, yyleng);
    #endif

    if (token_dump_file) {
      mode_buff.append( &mode_buff,
          yytext, yyleng );
    }
    sn_advance_column(yyleng);
  }
  [^\<\n]*\n {
    #if MATCH_DUMP
    matched_pattern("[^\\<\\n]*\\n", yytext);
    #endif

    #if HTML_DUMP
    fprintf(stderr, "html(2) \"%s\", %d\n", yytext, yyleng);
    #endif

    if (token_dump_file) {
      mode_buff.append( &mode_buff,
          yytext, yyleng );
    }
    sn_advance_line();
    sn_reset_column();
  }
  "<" {
    #if MATCH_DUMP
    matched_pattern("<", yytext);
    #endif

    #if HTML_DUMP
    fprintf(stderr, "html(3) \"%s\", %d\n", yytext, yyleng);
    #endif

    if (token_dump_file) {
      mode_buff.append( &mode_buff,
          yytext, yyleng );
    }
    sn_advance_column(yyleng);
  }
  ("<?"|"<?="|"<?php"|"<?PHP") {
    /* HTML -> PHP mode switch */
    #if MATCH_DUMP
    matched_pattern("(<?|<?=|<?php|<?PHP)", yytext);
    #endif
    sn_advance_column(yyleng);

    if (token_dump_file) {
      mode_buff.append( &mode_buff,
          yytext, yyleng );
    }

    emit_html();
    BEGIN(PHP);
  }
  /* Uncomment the next rule if you want to check to make sure
   * the above rules cover all possible input. A warning
   * "rule cannot be matched" should be printed by flex.
   */
  /*. {}*/
}

<PHP>("//"|"#").*("\n"|"?>") {
  char * x = (char *) yytext;
  char * y;
  int newline = 0;

#if MATCH_DUMP
  matched_pattern("(//|#).*\\n", yytext);
#endif

  while (*x == ' ' || *x == '\t') { x++; }
  assert(*x == '#' || (*x == '/' && *(x+1) == '/'));
  if (*x == '#') {
    x++;
  } else if (*x == '/') {
    x += 2;
  }
  sn_advance_column(x - yytext);

  /* Check for ?> in the text leading up to the newline.
   * PHP has this special rule where a line comment
   * can be terminated before the newline.
   */

  for (y=x; *y; y++) {
    if (*y == '?' && *(y+1) == '>') {
        int offset = (int) (y - yytext);
        yyless(offset); /* send everything after ?> back */
        break;
    }
  }

  y = yytext + yyleng - 1;
  if (*y == '\n') {
    newline = 1;
    *y = 0;
  }

  /* If dumping tokens, emit a special COMMENT token.
   * Otherwise, insert a comment symbol and a highlight.
   */
  if (token_dump_file) {
    append_token(COMMENT, x,
            sn_line(),
            sn_column() - (x - yytext),
            sn_line(),
            sn_column() + strlen(x));
  } else {
    sn_insert_comment(
      /* classname */ NULL,
      /* funcname */ NULL,
      sn_current_file(),
      x,
      sn_line(),
      sn_column());

    sn_highlight(SN_HIGH_COMMENT,
            sn_line(),
            sn_column() - (x - yytext),
            sn_line(),
            sn_column() + strlen(x));
  }

  if (newline) {
    sn_advance_line();
    sn_reset_column();
  } else {
    sn_advance_column(yyleng - (x - yytext));
  }
}

<PHP>"/*"		{ /* A C style multi-line comment, just like this! */
#if MATCH_DUMP
    matched_pattern("/*", yytext);
#endif
    BEGIN(COMMENT_MODE);
    sn_advance_column(2);
    LongStringInit(&mode_buff,0);
    mode_start_line = sn_line();
    mode_start_column = sn_column();
}

<COMMENT_MODE>{
  [^\*\n]* {
    #if MATCH_DUMP
    matched_pattern("[^\\*\\n]*", yytext);
    #endif

    #if COMMENT_DUMP
    fprintf(stderr, "comment(1) \"%s\", %d\n", yytext, yyleng);
    #endif
    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  [^\*\n]*\n {
    #if MATCH_DUMP
    matched_pattern("[^\\*\\n]*\\n", yytext);
    #endif

    #if COMMENT_DUMP
    fprintf(stderr, "comment(2) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_line();
    sn_reset_column();
  }
  \*+[^\*/\n]* {
    #if MATCH_DUMP
    matched_pattern("\\*+[^\\*/\\n]*", yytext);
    #endif

    #if COMMENT_DUMP
    fprintf(stderr, "comment(3) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  \*+[^\*/\n]*\n {
    #if MATCH_DUMP
    matched_pattern("\\*+[^\\*/\\n]*\\n", yytext);
    #endif

    #if COMMENT_DUMP
    fprintf(stderr, "comment(4) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_line();
    sn_reset_column();
  }
  "*"+"/" {
    #if MATCH_DUMP
    matched_pattern("\\*+/", yytext);
    #endif

    /* Include multiple '*' characters in the comment */
    if (yyleng > 2) {
        int to_append = yyleng;
        char *comment = (char *) yytext + yyleng - 1;
        assert(*comment == '/');
        comment--;
        assert(*comment == '*');
        *comment = '0';
        to_append -= 2;
        mode_buff.append( &mode_buff,
            yytext, to_append );
    }

    emit_comment();
    sn_advance_column(yyleng);
    BEGIN(PHP);
  }
  /* Uncomment the next rule if you want to check to make sure
   * the above rules cover all possible input. A warning
   * "rule cannot be matched" should be printed by flex.
   */
  /*. {}*/
}

<PHP>\"\" {
#if MATCH_DUMP
  matched_pattern("\\\"\\\"", yytext);
#endif
  /* FIXME: Can we pass NULL instead of "" after length issues worked out ? */
  append_dqstring_token("",
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>\" {
#if MATCH_DUMP
  matched_pattern("\\\"", yytext);
#endif

#if DQSTRING_DUMP
  fprintf(stderr, "dqstring started at (%d.%d)\n", sn_line(), sn_column());
#endif
  LongStringInit(&mode_buff,0);
  mode_start_line = sn_line();
  mode_start_column = sn_column();
  embedded_dq_string_vars_head = NULL;
  embedded_dq_string_vars_tail = NULL;
  sn_advance_column(yyleng);
  BEGIN(DQSTRING);
}

<DQSTRING>{
  [^\"\n\\\$]* {
    #if MATCH_DUMP
    matched_pattern("[^\\\"\\n\\\\]*", yytext);
    #endif

    #if DQSTRING_DUMP
    fprintf(stderr, "dqstring(1) \"%s\", %d\n", yytext, yyleng);
    #endif
    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  [^\"\n\\\$]*\n {
    #if MATCH_DUMP
    matched_pattern("[^\\\"\\n\\\\]*\\n", yytext);
    #endif

    #if DQSTRING_DUMP
    fprintf(stderr, "dqstring(2) \"%s\", %d\n", yytext, yyleng);
    #endif
    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_line();
    sn_reset_column();
  }
  (\\\")+ {
    #if MATCH_DUMP
    matched_pattern("(\\\")+", yytext);
    #endif

    #if DQSTRING_DUMP
    fprintf(stderr, "dqstring(3) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  \\ {
    #if MATCH_DUMP
    matched_pattern("\\", yytext);
    #endif

    #if DQSTRING_DUMP
    fprintf(stderr, "dqstring(4) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  \${varname} {
    #if MATCH_DUMP
    matched_pattern("${varname}", yytext);
    #endif

    append_dqstring_var_token(yytext+1,
            sn_line(),
            sn_column(),
            sn_line(),
            sn_column() + yyleng);

    mode_buff.append( &mode_buff,
        yytext, yyleng );

    sn_advance_column(yyleng);
  }
  \$ {
    #if MATCH_DUMP
    matched_pattern("$", yytext);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  \" {
    #if MATCH_DUMP
    matched_pattern("\\\"", yytext);
    #endif

    sn_advance_column(yyleng);
    emit_dqstring();
    BEGIN(PHP);
  }
  /* Uncomment the next rule if you want to check to make sure
   * the above rules cover all possible input. A warning
   * "rule cannot be matched" should be printed by flex.
   */
  /*. {}*/
}

<PHP>\'\' {
#if MATCH_DUMP
  matched_pattern("\\\'\\\'", yytext);
#endif

  /* FIXME: Can we pass NULL instead of "" after length issues worked out ? */
  append_token(SINGLE_QUOTED_STRING, "",
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>\' {
#if MATCH_DUMP
  matched_pattern("\\\'", yytext);
#endif

#if SQSTRING_DUMP
  fprintf(stderr, "sqstring started at (%d.%d)\n", sn_line(), sn_column());
#endif
  LongStringInit(&mode_buff,0);
  mode_start_line = sn_line();
  mode_start_column = sn_column();
  sn_advance_column(yyleng);
  BEGIN(SQSTRING);
}

<SQSTRING>{
  [^\'\n\\]* {
    #if MATCH_DUMP
    matched_pattern("[^\\\'\\n\\\\]*", yytext);
    #endif

    #if SQSTRING_DUMP
    fprintf(stderr, "sqstring(1) \"%s\", %d\n", yytext, yyleng);
    #endif
    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  [^\'\n\\]*\n {
    #if MATCH_DUMP
    matched_pattern("[^\\\'\\n\\\\]*\\n", yytext);
    #endif

    #if SQSTRING_DUMP
    fprintf(stderr, "sqstring(2) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_line();
    sn_reset_column();
  }
  (\\\')+[^\'/\n\\]* {
    #if MATCH_DUMP
    matched_pattern("\\\'+[^\\\'/\\n\\\\]*", yytext);
    #endif

    #if SQSTRING_DUMP
    fprintf(stderr, "sqstring(3) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  (\\\')+[^\'\n\\]*\n {
    #if MATCH_DUMP
    matched_pattern("\\\'+[^\\\'\\n\\\\]*\\n", yytext);
    #endif

    #if SQSTRING_DUMP
    fprintf(stderr, "sqstring(4) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_line();
    sn_reset_column();
  }
  \\ {
    #if MATCH_DUMP
    matched_pattern("\\\\", yytext);
    #endif

    #if SQSTRING_DUMP
    fprintf(stderr, "sqstring(5) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  \' {
    #if MATCH_DUMP
    matched_pattern("\\\'", yytext);
    #endif

    sn_advance_column(yyleng);
    emit_sqstring();
    BEGIN(PHP);
  }
  /* Uncomment the next rule if you want to check to make sure
   * the above rules cover all possible input. A warning
   * "rule cannot be matched" should be printed by flex.
   */
  /*. {}*/
}

<PHP>"<<<"{ws}*{heredoc-id}\n {
  char *x = yytext;
  char *y;

#if MATCH_DUMP
  matched_pattern("<<<{ws}*{heredoc-id}\\n", yytext);
#endif

  mode_start_line = sn_line();
  mode_start_column = sn_column();

  x += 3;
  while(*x == ' ' || *x == '\t') { x++; }
  y = x;
  while(*y != '\n') { y++; }
  *y = 0;

  LongStringInit(&mode_buff,0);
  mode_buff.append( &mode_buff,
          x, -1 );
  mode_buff.append( &mode_buff,
          ";\n", -1 );
  heredoc_id = SN_StrDup(mode_buff.buf);
  mode_buff.free(&mode_buff);

#if HDSTRING_DUMP
  fprintf(stderr, "hdstring started at (%d.%d)\n", sn_line(), sn_column());
  fprintf(stderr, "hdstring id is \"%s\"\n", heredoc_id);
#endif

  LongStringInit(&mode_buff,0);

  embedded_dq_string_vars_head = NULL;
  embedded_dq_string_vars_tail = NULL;

  sn_advance_line();
  sn_reset_column();

  BEGIN(HDSTRING);
}

<HDSTRING>{
  ^{heredoc-id};\n {
    #if MATCH_DUMP
    matched_pattern("^{heredoc-id};\n", yytext);
    #endif

    #if HDSTRING_DUMP
    fprintf(stderr, "hdstring(1) \"%s\", %d\n", yytext, yyleng);
    #endif

    if (strcmp(yytext, heredoc_id) == 0) {
        #if HDSTRING_DUMP
        fprintf(stderr, "hdstring end matched\n");
        #endif

        sn_advance_column(yyleng-2); /* advance to ; */
        emit_hdstring();
        sn_advance_line();
        sn_reset_column();
        BEGIN(PHP);
    } else {
        mode_buff.append( &mode_buff, yytext, yyleng );
    }
  }
  [^\$\n]+ {
    #if MATCH_DUMP
    matched_pattern("[^;$\n]+", yytext);
    #endif

    #if HDSTRING_DUMP
    fprintf(stderr, "hdstring(2) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );

    sn_advance_column(yyleng);
  }
  [^\$\n]*\n {
    #if MATCH_DUMP
    matched_pattern("[^;$\n]*\n", yytext);
    #endif

    #if HDSTRING_DUMP
    fprintf(stderr, "hdstring(3) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );

    sn_advance_line();
    sn_reset_column();
  }
  \${varname} {
    #if MATCH_DUMP
    matched_pattern("${varname}", yytext);
    #endif

    #if HDSTRING_DUMP
    fprintf(stderr, "hdstring(4) \"%s\", %d\n", yytext, yyleng);
    #endif

    append_dqstring_var_token(yytext+1,
            sn_line(),
            sn_column(),
            sn_line(),
            sn_column() + yyleng);

    mode_buff.append( &mode_buff,
        yytext, yyleng );

    sn_advance_column(yyleng);
  }
  \$ {
    #if MATCH_DUMP
    matched_pattern("$", yytext);
    #endif

    #if HDSTRING_DUMP
    fprintf(stderr, "hdstring(5) \"%s\", %d\n", yytext, yyleng);
    #endif

    mode_buff.append( &mode_buff,
        yytext, yyleng );
    sn_advance_column(yyleng);
  }
  /* Uncomment the next rule if you want to check to make sure
   * the above rules cover all possible input. A warning
   * "rule cannot be matched" should be printed by flex.
   */
  /*. {}*/
}

<PHP>"(" {
#if MATCH_DUMP
  matched_pattern("(", yytext);
#endif

  append_token(OPEN_PAREN, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>")" {
#if MATCH_DUMP
  matched_pattern(")", yytext);
#endif
  append_token(CLOSE_PAREN, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"[" {
#if MATCH_DUMP
  matched_pattern("[", yytext);
#endif
  append_token(OPEN_BRACKET, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"]" {
#if MATCH_DUMP
  matched_pattern("]", yytext);
#endif
  append_token(CLOSE_BRACKET, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"{" {
#if MATCH_DUMP
  matched_pattern("{", yytext);
#endif
  append_token(OPEN_BRACE, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"}" {
#if MATCH_DUMP
  matched_pattern("}", yytext);
#endif
  append_token(CLOSE_BRACE, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>";" {
#if MATCH_DUMP
  matched_pattern("}", yytext);
#endif
  append_token(SEMICOLON, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"." {
#if MATCH_DUMP
  matched_pattern(".", yytext);
#endif
  append_token(PERIOD, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"," {
#if MATCH_DUMP
  matched_pattern(",", yytext);
#endif
  append_token(COMMA, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"=" {
#if MATCH_DUMP
  matched_pattern("=", yytext);
#endif
  append_token(ASSIGNMENT_OPERATOR, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"&" {
#if MATCH_DUMP
  matched_pattern("&", yytext);
#endif
  append_token(REFERENCE_OPERATOR, NULL,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{comparison-operators} {
#if MATCH_DUMP
  matched_pattern("{comparison-operators}", yytext);
#endif

  append_token(COMPARISON_OPERATORS, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{assignment-operators} {
#if MATCH_DUMP
  matched_pattern("{assignment-operators}", yytext);
#endif

  append_token(ASSIGNMENT_OPERATORS, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{increment-operators} {
#if MATCH_DUMP
  matched_pattern("{increment-operators}", yytext);
#endif

  append_token(INCREMENT_OPERATORS, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"function" {
#if MATCH_DUMP
  matched_pattern("function", yytext);
#endif

  append_token(FUNCTION_KEYWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>"global" {
#if MATCH_DUMP
  matched_pattern("global", yytext);
#endif
  append_token(GLOBAL_KEYWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{include-keywords} {
#if MATCH_DUMP
  matched_pattern("{include-keywords}", yytext);
#endif
  append_token(INCLUDE_KEYWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{keywords} {
#if MATCH_DUMP
  matched_pattern("{keywords}", yytext);
#endif

  append_token(KEYWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{someword} {
#if MATCH_DUMP
  matched_pattern("{someword}", yytext);
#endif

  append_token(SOMEWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>\${varname} {
#if MATCH_DUMP
  matched_pattern("${varname}", yytext);
#endif
  append_token(VARIABLE, yytext+1,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);
  sn_advance_column(yyleng);
}

<PHP>{wsn}+ {
  char* x, *y;
  #if MATCH_DUMP
  matched_pattern("{wsn}+", yytext);
  #endif

  for (x=yytext, y=NULL; *x ; x++) {
    if (*x == '\n') {
        y=x+1;
        sn_advance_line();
    }
  }
  if (y == NULL) {
    sn_advance_column(yyleng);
  } else {
    sn_reset_column();
    sn_advance_column(x-y);
  }
}

<PHP>\\\$		{
#if MATCH_DUMP
  matched_pattern("\\$", yytext);
#endif

  append_token(SOMEWORD, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);

  sn_advance_column(yyleng);
}

<PHP>.		{
#if MATCH_DUMP
  matched_pattern(".", yytext);
#endif

  /* Add an UNKNOWN token for each
   * character that we don't know
   * how to deal with.
   */

  append_token(UNKNOWN, yytext,
          sn_line(),
          sn_column(),
          sn_line(),
          sn_column() + yyleng);

  /*fprintf(stderr, "adding unknown token for \"%s\"\n", yytext);*/

  sn_advance_column(yyleng); /* eat text */
}



<TOKEN>"FUNCTION_KEYWORD SOMEWORD OPEN_PAREN"{function-argument-tokens}"CLOSE_PAREN OPEN_BRACE " {
  int parens, noargs;
  LongString abuff;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "function declaration at token %d\n", token_index);
  fprintf(tokenout, "match text was \"%s\"\n", yytext);
#endif
  assert(tokens_head->type == FUNCTION_KEYWORD);

  current_function_line_start = tokens_head->start_line;
  current_function_column_start = tokens_head->start_column;

  sn_highlight(SN_HIGH_KEYWORD,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);

  free_head_token(); /* FUNCTION_KEYWORD */

  if (current_function != NULL) {
    push_function();
  }
  current_function = SN_StrDup(tokens_head->strval);

  current_function_highlight_line = tokens_head->start_line;
  current_function_highlight_column_start = tokens_head->start_column;
  current_function_highlight_column_end = tokens_head->end_column;

  free_head_token(); /* SOMEWORD */
  free_head_token(); /* OPEN_PAREN */

  /* Save function arguments in current_function_args */  
  parens = 1;
  noargs = 1;
  LongStringInit(&abuff,0);
  while (parens > 0) {
    if (tokens_head->type == OPEN_PAREN) {
      parens++;
    } else if (tokens_head->type == CLOSE_PAREN) {
      parens--;
    } else if (tokens_head->type == VARIABLE) {
        /* We just ignore default values for arguments */
        if (!noargs) {
            abuff.append( &abuff, ",", -1);
        }
        abuff.append( &abuff, tokens_head->strval, -1);
        noargs = 0;

        sn_highlight(SN_HIGH_VAR_LOCAL,
            tokens_head->start_line, tokens_head->start_column,
            tokens_head->end_line, tokens_head->end_column);
    } else if ((tokens_head->type == DOUBLE_QUOTED_STRING) ||
               (tokens_head->type == SINGLE_QUOTED_STRING)) {
        sn_highlight(SN_HIGH_STRING,
            tokens_head->start_line, tokens_head->start_column,
            tokens_head->end_line, tokens_head->end_column);
    }

    free_head_token(); /* OPEN_PAREN || CLOSE_PAREN || VARIABLE || COMMA */
  }

  assert(current_function_args == NULL);
  if (!noargs) {
      current_function_args = SN_StrDup(abuff.buf);
  }
  abuff.free(&abuff);

  free_head_token(); /* OPEN_BRACE */

  current_function_brace_count = 1;
}

<TOKEN>"OPEN_BRACE" {
  if (current_function) {
    current_function_brace_count++;
  }
  free_head_token(); /* OPEN_BRACE */
}

<TOKEN>"CLOSE_BRACE" {
  if (current_function && (--current_function_brace_count == 0)) {
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "end of function %s at token %d\n",
        current_function, token_index);
#endif

    current_function_line_end = tokens_head->end_line;
    current_function_column_end = tokens_head->end_column;

    emit_function_declaration();

    pop_function(); /* in case of nested function */
  }

  free_head_token(); /* CLOSE_BRACE */
}

<TOKEN>("INCLUDE_KEYWORD "{include-string-token}" SEMICOLON")|("INCLUDE_KEYWORD OPEN_PAREN "{include-string-token}" CLOSE_PAREN") {
  int line_start, line_end, column_start, column_end;
  char * filename;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "include statement begins at token %d\n", token_index);
#endif

  sn_highlight(SN_HIGH_KEYWORD,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);

  free_head_token(); /* INCLUDE_KEYWORD */

  if (tokens_head->type == OPEN_PAREN) {
    free_head_token(); /* OPEN_PAREN */
  }

  assert((tokens_head->type == SINGLE_QUOTED_STRING) ||
           (tokens_head->type == DOUBLE_QUOTED_STRING));

  line_start = tokens_head->start_line;
  column_start = tokens_head->start_column + 1;
  line_end = tokens_head->end_line;
  column_end = tokens_head->end_column - 1;

  filename = tokens_head->strval;

  result = sn_insert_symbol(SN_INCLUDE_DEF,
          /* classname */ NULL,
          filename,
          sn_current_file(),
          line_start, column_start,
          line_end, column_end,
          /* attr */ 0,
          /* ret */ NULL,
          /* arg_types */ NULL,
          /* arg_names */ NULL,
          /* comment */ NULL,
          line_start, column_start,
          line_end, column_end);

  assert(result == 0);

  sn_highlight(SN_HIGH_STRING,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);

  free_head_token(); /* SINGLE_QUOTED_STRING || DOUBLE_QUOTED_STRING */
  free_head_token(); /* CLOSE_PAREN || SEMICOLON */
}

<TOKEN>("GLOBAL_KEYWORD"{variable-tokens}"SEMICOLON") {
  SearchEntry entry;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "global statement found at %d\n", token_index);
#endif

  sn_highlight(SN_HIGH_KEYWORD,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);

  free_head_token(); /* GLOBAL_KEYWORD */

  if (current_function && (global_var_table == NULL)) {
    global_var_table = SearchTableCreate(100, SEARCH_HASH_TABLE, FreeGlobalEntry);
  }

  /* Insert the variable name into the function's global table */

  while (tokens_head->type != SEMICOLON) {
    if (current_function && (tokens_head->type == VARIABLE)) {
        entry.key = tokens_head->strval;
        entry.key_len = -1;

        if (global_var_table->search( &global_var_table, entry ) == NULL)
        {
            entry.data = NULL;
            entry.data_len = 0;
            entry.flag = SEARCH_DUP_KEY; /* add copy of entry.key */
            global_var_table->add( &global_var_table, entry );

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "added global \"%s\"\n", entry.key);
#endif
        }

        sn_highlight(SN_HIGH_VAR_GLOBAL,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);
    }

    free_head_token(); /* VARIABLE || COMMA */
  }

  free_head_token(); /* SEMICOLON */
}

<TOKEN>"SOMEWORD OPEN_PAREN" {
  char* fname;
  int line;
  int ref_from_scope_type;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "found function call tokens at %d\n", token_index);
  fprintf(tokenout, "match text was \"%s\"\n", yytext);
#endif

  fname = tokens_head->strval;
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "function name is \"%s\"\n", fname);
#endif
  line = tokens_head->start_line;

  /*
   * Pass SN_GLOBAL_NAMESPACE if the xref is outside of
   * a function, otherwise pass SN_FUNC_DEF.
   */   

  if (current_function == NULL) {
    ref_from_scope_type = SN_GLOBAL_NAMESPACE;
  } else {
    ref_from_scope_type = SN_FUNC_DEF;
  }

  result = sn_insert_xref(SN_REF_TO_FUNCTION,
                 ref_from_scope_type,
                 SN_REF_SCOPE_GLOBAL,
                 NULL,
                 current_function,
                 NULL,
                 NULL,
                 fname,
                 NULL, 
                 sn_current_file(),
                 line,
                 SN_REF_PASS);

  assert(result == 0);

  sn_highlight(SN_HIGH_FUNCTION,
      tokens_head->start_line, tokens_head->start_column,
      tokens_head->end_line, tokens_head->end_column);

  free_head_token(); /* SOMEWORD */
  free_head_token(); /* OPEN_PAREN */
}

<TOKEN>"VARIABLE ASSIGNMENT_OPERATOR" {
  int offset;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "variable assignment at token %d\n", token_index);
#endif

  emit_var_access(tokens_head, VAR_WRITE);

  free_head_token(); /* VARIABLE */
  free_head_token(); /* ASSIGNMENT_OPERATOR */
}

<TOKEN>"VARIABLE ASSIGNMENT_OPERATORS" {
  int offset;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "variable modification assignment at token %d\n", token_index);
#endif

  emit_var_access(tokens_head, VAR_READWRITE);

  free_head_token(); /* VARIABLE */
  free_head_token(); /* ASSIGNMENT_OPERATORS */
}

<TOKEN>("INCREMENT_OPERATORS VARIABLE"|"VARIABLE INCREMENT_OPERATORS") {
  int offset, pre = 0;
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "variable increment at token %d\n", token_index);
#endif

  if (tokens_head->type == INCREMENT_OPERATORS) {
    pre = 1;
    free_head_token(); /* INCREMENT_OPERATORS */
  }

  emit_var_access(tokens_head, VAR_READWRITE);

  free_head_token(); /* VARIABLE */
  if (!pre) {
    free_head_token(); /* INCREMENT_OPERATORS */
  }
}

<TOKEN>"VARIABLE" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "variable read at token %d\n", token_index);
#endif

  emit_var_access(tokens_head, VAR_READ);

  free_head_token(); /* VARIABLE */
}

<TOKEN>"INCREMENT_OPERATORS VARIABLE OPEN_BRACKET" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "increment array variable start token %d\n", token_index);
#endif

  /* This is the start of a prefix incremented array variable
   *
   *   ++$arr[0]
   */

  if (current_array_bracket_count > 0) {
    /* Nested brackets : $arr1[++$arr2[0]] */
    push_bracket_counter();
  }
  current_array_bracket_count = 1;
  current_array_prefix_incr = 1;

  free_head_token(); /* INCREMENT_OPERATORS */
  current_array_token = pop_head_token(); /* VARIABLE */
  free_head_token(); /* OPEN_BRACKET */
}

<TOKEN>"VARIABLE OPEN_BRACKET" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable start token %d\n", token_index);
#endif

  /* This is the start of an array variable.
   *
   *   $arr[0]
   */

  if (current_array_bracket_count > 0) {
    /* Nested brackets : $arr1[$arr2[0]] */
    push_bracket_counter();
  }
  current_array_bracket_count = 1;
  current_array_token = pop_head_token(); /* VARIABLE */
  free_head_token(); /* OPEN_BRACKET */
}

<TOKEN>"CLOSE_BRACKET OPEN_BRACKET" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable array dimension tokens %d\n", token_index);
#endif

  /* This rule matches the "][" tokens in the following statement:
   *
   * array["one"]["two"] = 1;
   *
   * Note that we don't change the current_array_bracket_count here.
   */

  free_head_token(); /* CLOSE_BRACKET */
  free_head_token(); /* OPEN_BRACKET */
}

<TOKEN>"CLOSE_BRACKET ASSIGNMENT_OPERATOR" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable assignment end token %d\n", token_index);
#endif

  /* This could be the end of an array variable.
   *
   *   $arr["num"] = 2;
   */

  if (current_array_bracket_count > 0) {
    if (--current_array_bracket_count == 0) {
        emit_var_access(current_array_token, VAR_WRITE);
        FreeToken(current_array_token);
        current_array_prefix_incr = 0;

        pop_bracket_counter();

#ifdef TOKEN_DEBUG
        fprintf(tokenout, "assignment operator end of array variable\n");
#endif
    }
  }

  free_head_token(); /* CLOSE_BRACKET */
  free_head_token(); /* ASSIGNMENT_OPERATOR */
}

<TOKEN>"CLOSE_BRACKET ASSIGNMENT_OPERATORS" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable read/write assignment end token %d\n", token_index);
#endif

  /* This could be the end of an array variable.
   *
   *   $arr["num"] += 2;
   */

  if (current_array_bracket_count > 0) {
    if (--current_array_bracket_count == 0) {
        emit_var_access(current_array_token, VAR_READWRITE);
        FreeToken(current_array_token);
        current_array_prefix_incr = 0;
        pop_bracket_counter();

#ifdef TOKEN_DEBUG
        fprintf(tokenout, "assignment operators end of array variable\n");
#endif
    }
  }

  free_head_token(); /* CLOSE_BRACKET */
  free_head_token(); /* ASSIGNMENT_OPERATORS */
}

<TOKEN>"CLOSE_BRACKET INCREMENT_OPERATORS" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable post increment end token %d\n", token_index);
#endif

  /* This could be the end of an array variable increment
   *
   *   $arr[0]++
   */

  if (current_array_bracket_count > 0) {
    if (--current_array_bracket_count == 0) {
        emit_var_access(current_array_token, VAR_READWRITE);
        current_array_prefix_incr = 0;
        FreeToken(current_array_token);

        pop_bracket_counter();

#ifdef TOKEN_DEBUG
        fprintf(tokenout, "end of array variable post increment\n");
#endif
    }
  }

  free_head_token(); /* CLOSE_BRACKET */
  free_head_token(); /* INCREMENT_OPERATORS */
}

<TOKEN>"CLOSE_BRACKET" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "array variable read end token %d\n", token_index);
#endif

  /* This could be the end of an array variable
   *
   *   $arr[0] = 2;
   *
   * Or the end of an array variable that has
   * a prefix increment.
   *
   *   ++$arr[0]
   */

  if (current_array_bracket_count > 0) {
    if (--current_array_bracket_count == 0) {
        if (current_array_prefix_incr) {
          emit_var_access(current_array_token, VAR_READWRITE);
#ifdef TOKEN_DEBUG
        fprintf(tokenout, "read/write end of array variable\n");
#endif
        } else {
          emit_var_access(current_array_token, VAR_READ);
#ifdef TOKEN_DEBUG
        fprintf(tokenout, "read end of array variable\n");
#endif
        }
        current_array_prefix_incr = 0;
        FreeToken(current_array_token);

        pop_bracket_counter();
    }
  }

  free_head_token(); /* CLOSE_BRACKET */
}

<TOKEN>"VDOUBLE_QUOTED_STRING" {
#ifdef TOKEN_DEBUG
  fprintf(tokenout, "ate VDOUBLE_QUOTED_STRING token %d", token_index);
#endif

  sn_highlight(SN_HIGH_STRING,
          tokens_head->start_line, tokens_head->start_column,
          tokens_head->end_line, tokens_head->end_column);

  emit_var_access_for_dqstring(tokens_head);

  free_head_token(); /* VDOUBLE_QUOTED_STRING */
}

<TOKEN>{token} {
  enum sn_highlights type;

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "ate token %d %s", token_index,
          TokenTypeToString(tokens_head));
  if (tokens_head->strval) {
      fprintf(tokenout, " \"%s\"", tokens_head->strval);
  }
  fprintf(tokenout, "\n");
#endif

  if (highlight_file) {
      switch (tokens_head->type) {
          case DOUBLE_QUOTED_STRING:
          case SINGLE_QUOTED_STRING:
              type = SN_HIGH_STRING;
              break;
          case KEYWORD:
              type = SN_HIGH_KEYWORD;
              break;
          default:
              type = 0;
      }
      if (type != 0) {
          sn_highlight(type,
                  tokens_head->start_line, tokens_head->start_column,
                  tokens_head->end_line, tokens_head->end_column);
      }
  }

  free_head_token(); /* ... */
}

<TOKEN>" "

<TOKEN>.		{
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "matched unknown character \"%s\"\n", yytext);
#endif
}

<TOKEN><<EOF>> {
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "reached EOF in TOKEN buffer\n");
#endif

    /* A function closing brace was not found before we hit EOF */
    if (current_function) {
        current_function_line_end = sn_line();
        current_function_column_end = sn_column();

#ifdef TOKEN_DEBUG
        fprintf(tokenout, "found unfinished function at EOF in %s\n", sn_current_file());
#endif

        emit_function_declaration();
    }

    assert(!tokens_head); /* all tokens were processed */

    yy_delete_buffer( YY_CURRENT_BUFFER );
    yy_switch_to_buffer( original_buffer );
    enter_html_mode();
    yyterminate();
}

<HTML_MODE,PHP,COMMENT_MODE,DQSTRING,SQSTRING,HDSTRING><<EOF>> {
  LongString token_buffer;
  char *base;
  int i;
  Token* tok;
  yy_size_t size;
  YY_BUFFER_STATE yybs;

#ifdef TOKEN_DEBUG
  tokenout = fopen("tokens.out", "a");
  fprintf(tokenout, "reached EOF in lex input buffer in mode %s\n", modestring());
#endif

  /* See if we ran off the end of the lex input buffer in a special mode */
  switch (YY_START) {
    case COMMENT_MODE:
      emit_comment();
      break;
    case DQSTRING:
      emit_dqstring();
      break;
    case SQSTRING:
      emit_sqstring();
      break;
    case HDSTRING: 
      emit_hdstring();
      break;
    case HTML_MODE:
      emit_html();
      break;
  }

  /* If no tokens were generated, then quit now */
  if (tokens_head == NULL) {
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "no TOKENs generated\n");
#endif
    enter_html_mode();
    yyterminate();
  }

  /*
   * If the -T command line option was passed,
   * dump all tokens to a file, skip the token
   * matching phase and go on to the next file.
   */

  if (token_dump_file) {
    FILE * dump_tokens = fopen(token_dump_file, "a");

    for (i=0, tok = tokens_head ; tok ; tok = tok->next, i++) {
        fprintf(dump_tokens, "%d %s", i, TokenTypeToString(tok));
        if (tok->strval == NULL) {
            fprintf(dump_tokens, " \"\"");
        } else {
            char *x;
            fprintf(dump_tokens, " \"");
            for (x=tok->strval; *x; x++) {
                if (*x == '\n') {
                    fprintf(dump_tokens, "\\n");
                } else if (*x == '\\') {
                    fprintf(dump_tokens, "\\\\");
                } else if (*x == '\"') {
                    fprintf(dump_tokens, "\\\"");
                } else {
                    fprintf(dump_tokens, "%c", *x);
                }
            }
            fprintf(dump_tokens, "\"");
        }
        fprintf(dump_tokens, " %d.%d %d.%d",
            tok->start_line,
            tok->start_column,
            tok->end_line,
            tok->end_column
        );
        fprintf(dump_tokens, "\n");
    }

    fclose(dump_tokens);

    enter_html_mode();
    yyterminate();
  }

  LongStringInit(&token_buffer,0);

  /* Print token info to in memory buffer and then reload
     the input state machine and start out in the TOKEN mode. */

  for (i=0, tok = tokens_head ; tok ; tok = tok->next, i++) {
#ifdef TOKEN_DEBUG
        fprintf(tokenout, "token %d %s", i, TokenTypeToString(tok));
        if (tok->strval) {
            fprintf(tokenout, " \"%s\"", tok->strval);
        }
        fprintf(tokenout, " (%d.%d -> %d.%d)",
            tok->start_line,
            tok->start_column,
            tok->end_line,
            tok->end_column
        );
        fprintf(tokenout, "\n");
#endif
        
        token_buffer.append( &token_buffer, TokenTypeToString(tok), -1);
        token_buffer.append( &token_buffer, " ", -1);
  }

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "token buffer data is \"%s\"\n", token_buffer.buf);
#endif

  original_buffer = YY_CURRENT_BUFFER;
  yy_switch_to_buffer( yy_scan_string(token_buffer.buf) );

  token_buffer.free(&token_buffer);

#ifdef TOKEN_DEBUG
  fprintf(tokenout, "switching to token mode\n");
#endif

  BEGIN(TOKEN);
}

%%

/* Return a string that describes the current mode */

static char* modestring() {
    char* mode = "UNKNOWN";

    switch (YY_START) {
        case INITIAL:
            mode = "INITIAL";
            break;
        case PHP:
            mode = "PHP";
            break;
        case COMMENT_MODE:
            mode = "COMMENT_MODE";
            break;
        case DQSTRING:
            mode = "DQSTRING";
            break;
        case SQSTRING:
            mode = "SQSTRING";
            break;
        case HDSTRING:
            mode = "HDSTRING";
            break;
        case TOKEN:
            mode = "TOKEN";
            break;
        case HTML_MODE:
            mode = "HTML_MODE";
            break;
    }
    return mode;
}

#if MATCH_DUMP

/* Helper method that will print matches as they are made.
 * This method is typically used in the token generation phase.
 */

static void matched_pattern(char * pattern, char * text) {
    char * mode = modestring();

    fprintf(stderr, "Matched \"%s\", with text \"%s\", in mode \"%s\" (%d.%d)\n",
        pattern, text, mode, sn_line(), sn_column());
}

#endif /* MATCH_DUMP */

static void FreeGlobalEntry(SearchEntry *entry) {}


void FreeToken(Token* tok) {
    if (tok->strval != NULL) {
        ckfree(tok->strval);
    }
    if (tok->vars != NULL) {
        Token* tmp, *next;
        for (tmp=tok->vars; tmp; ) {
            next = tmp->next;
            FreeToken(tmp);
            tmp = next;
        }
    }
    ckfree((char *) tok);
}

void append_dqstring_var_token(char* var,
                           long start_line,
                           int start_column,
                           long end_line,
                           int end_column) {
    Token* tok;
    tok = (Token*) ckalloc(sizeof(Token));
    tok->type = VARIABLE;
    tok->strval = SN_StrDup(var);

    tok->start_line = start_line;
    tok->start_column = start_column;

    tok->end_line = end_line;
    tok->end_column = end_column;
    tok->vars = NULL;
    tok->next = NULL;

    /* append to var token list */

    if (embedded_dq_string_vars_tail == NULL) {
        embedded_dq_string_vars_head = embedded_dq_string_vars_tail = tok;
    } else {
        embedded_dq_string_vars_tail->next = tok;
        embedded_dq_string_vars_tail = tok;
    }
}

void append_dqstring_token(char* strval,
                           long start_line,
                           int start_column,
                           long end_line,
                           int end_column) {
    if (embedded_dq_string_vars_head != NULL) {
        append_token(VDOUBLE_QUOTED_STRING,
                     strval,
                     start_line,
                     start_column,
                     end_line,
                     end_column);

        tokens_tail->vars = embedded_dq_string_vars_head;
        embedded_dq_string_vars_head = NULL;
        embedded_dq_string_vars_tail = NULL;
    } else {
        append_token(DOUBLE_QUOTED_STRING,
                     strval,
                     start_line,
                     start_column,
                     end_line,
                     end_column);
    }
}

void append_token(TokenType type,
                  char* strval,
                  long start_line,
                  int start_column,
                  long end_line,
                  int end_column) {
    Token* tok;
    tok = (Token*) ckalloc(sizeof(Token));
    tok->type = type;
    if (strval)
        tok->strval = SN_StrDup(strval);
    else
        tok->strval = NULL;

    tok->start_line = start_line;
    tok->start_column = start_column;

    tok->end_line = end_line;
    tok->end_column = end_column;
    tok->vars = NULL;
    tok->next = NULL;

    /* append to token list */

    if (tokens_tail == NULL) {
        tokens_head = tokens_tail = tok;
    } else {
        tokens_tail->next = tok;
        tokens_tail = tok;
    }
}

Token* pop_head_token() {
    Token* tok;
    assert(tokens_head);
    tok = tokens_head;
    if (tokens_head == tokens_tail) {
        tokens_head = tokens_tail = (Token*) NULL;
    } else {
        tokens_head = tokens_head->next;
    }
    token_index++;
    return tok;
}

void free_head_token() {
    FreeToken(pop_head_token());
}

/*
 * This method is invoked when nested brackets within an
 * array name are encountered. For example: $arr1[++$arr2[0]]
 */

void push_bracket_counter() {
    NestedBracket* nbp = (NestedBracket*) ckalloc(sizeof(NestedBracket));
    nbp->var = current_array_token;
    nbp->count = current_array_bracket_count;
    nbp->prefix_incr = current_array_prefix_incr;

    if (array_nest_head == NULL) {
        array_nest_head = array_nest_tail = nbp;
    } else {
        nbp->next = array_nest_head;
        array_nest_head = nbp;
    }
}

void pop_bracket_counter() {
    if (array_nest_head != NULL) {
        NestedBracket* nbp = array_nest_head;
        if (array_nest_head == array_nest_tail) {
            array_nest_head = array_nest_tail = (NestedBracket*) NULL;
        } else {
            array_nest_head = array_nest_head->next;
        }
        current_array_token = nbp->var;
        current_array_bracket_count = nbp->count;
        current_array_prefix_incr = nbp->prefix_incr;
        ckfree((char*) nbp);
    }
}

/*
 * This method is invoked when a nested function is
 * encountered.
 */

void push_function() {
    NestedFunction* n = (NestedFunction*) ckalloc(sizeof(NestedFunction));

    n->name = current_function;
    n->args = current_function_args;
    n->high_line = current_function_highlight_line;
    n->high_col_start = current_function_highlight_column_start;
    n->high_col_end = current_function_highlight_column_end;
    n->line_start = current_function_line_start;
    n->line_end = current_function_line_end;
    n->col_start = current_function_column_start;
    n->col_end = current_function_column_end;
    n->brace_count = current_function_brace_count;

    if (function_nest_head == NULL) {
        function_nest_head = function_nest_tail = n;
    } else {
        n->next = function_nest_head;
        function_nest_head = n;
    }
}

void pop_function() {
    if (function_nest_head != NULL) {
        NestedFunction* n = function_nest_head;
        if (function_nest_head == function_nest_tail) {
            function_nest_head = function_nest_tail = (NestedFunction*) NULL;
        } else {
            function_nest_head = function_nest_head->next;
        }

        current_function = n->name;
        current_function_args = n->args;
        current_function_highlight_line = n->high_line;
        current_function_highlight_column_start = n->high_col_start;
        current_function_highlight_column_end = n->high_col_end;
        current_function_line_start = n->line_start;
        current_function_line_end = n->line_end;
        current_function_column_start = n->col_start;
        current_function_column_end = n->col_end;
        current_function_brace_count = n->brace_count;

        ckfree((char*) n);
    }
}

char * TokenTypeToString(Token *tok) {
    switch(tok->type) {
        case OPEN_PAREN:
            return "OPEN_PAREN";
        case CLOSE_PAREN:
            return "CLOSE_PAREN";
        case OPEN_BRACE:
            return "OPEN_BRACE";
        case CLOSE_BRACE:
            return "CLOSE_BRACE";
        case OPEN_BRACKET:
            return "OPEN_BRACKET";
        case CLOSE_BRACKET:
            return "CLOSE_BRACKET";
        case SEMICOLON:
            return "SEMICOLON";
        case COMMA:
            return "COMMA";
        case PERIOD:
            return "PERIOD";
        case VARIABLE:
            return "VARIABLE";
        case ASSIGNMENT_OPERATOR:
            return "ASSIGNMENT_OPERATOR";
        case ASSIGNMENT_OPERATORS:
            return "ASSIGNMENT_OPERATORS";
        case COMPARISON_OPERATORS:
            return "COMPARISON_OPERATORS";
        case INCREMENT_OPERATORS:
            return "INCREMENT_OPERATORS";
        case REFERENCE_OPERATOR:
            return "REFERENCE_OPERATOR";
        case SOMEWORD:
            return "SOMEWORD";
        case KEYWORD:
            return "KEYWORD";
        case FUNCTION_KEYWORD:
            return "FUNCTION_KEYWORD";
        case GLOBAL_KEYWORD:
            return "GLOBAL_KEYWORD";
        case INCLUDE_KEYWORD:
            return "INCLUDE_KEYWORD";
        case VDOUBLE_QUOTED_STRING:
            return "VDOUBLE_QUOTED_STRING";
        case DOUBLE_QUOTED_STRING:
            return "DOUBLE_QUOTED_STRING";
        case SINGLE_QUOTED_STRING:
            return "SINGLE_QUOTED_STRING";
        case UNKNOWN:
            return "UNKNOWN";
        case COMMENT:
            return "COMMENT";
        case HTML:
            return "HTML";
        default:
            return "TOKEN_NOT_MATCHED";
    }
}

/* Called when the closing brace of a function is found
 * or when we hit EOF without finding the end of the
 * function.
 */

void emit_function_declaration() {
    result = sn_insert_symbol(SN_FUNC_DEF,
            NULL,
            current_function,
            sn_current_file(), 
            current_function_line_start, current_function_column_start,
            current_function_line_end, current_function_column_end,
            0 /* attribute */,
            NULL /* return type */,
            NULL /* argument types */,
            current_function_args /* argument names */,
            NULL /* comment */,
            current_function_highlight_line,
            current_function_highlight_column_start,
            current_function_highlight_line,
            current_function_highlight_column_end );

    assert(result == 0);

    ckfree(current_function);
    current_function = NULL;
    ckfree(current_function_args);
    current_function_args = NULL;

    if (global_var_table) {
      global_var_table->destroy( &global_var_table );
      global_var_table = NULL;
    }
}

void emit_comment() {
    char* comment = mode_buff.buf;

#if COMMENT_DUMP
    fprintf(stderr, "emit comment \"%s\"\n", comment);
#endif

  /* If dumping tokens, emit a special COMMENT token.
   * Otherwise, insert a comment symbol and a highlight.
   */
  if (token_dump_file) {
    append_token(COMMENT, comment,
            mode_start_line,
            mode_start_column - 2,
            sn_line(),
            sn_column() + 2);
  } else {
    sn_insert_comment(
        /* classname */ NULL,
        /* funcname */ NULL,
        sn_current_file(),
        comment,
        mode_start_line,
        mode_start_column);

    sn_highlight(SN_HIGH_COMMENT,
            mode_start_line,
            mode_start_column - 2,
            sn_line(),
            sn_column() + 2);
  }

  mode_buff.free(&mode_buff);
}

void emit_dqstring() {
    char* dqstring = mode_buff.buf;
    char * x;
    char * var;

#if DQSTRING_DUMP
    fprintf(stderr, "creating dqstring token \"%s\"\n", dqstring);
#endif

    append_dqstring_token(dqstring,
            mode_start_line,
            mode_start_column,
            sn_line(),
            sn_column());

    mode_buff.free(&mode_buff);
}

void emit_sqstring() {
    char* sqstring = mode_buff.buf;

#if SQSTRING_DUMP
    fprintf(stderr, "creating sqstring token \'%s\'\n", sqstring);
#endif

    append_token(SINGLE_QUOTED_STRING, sqstring,
            mode_start_line,
            mode_start_column,
            sn_line(),
            sn_column());

    mode_buff.free(&mode_buff);
}

void emit_hdstring() {
    char* hdstring = mode_buff.buf;

    /* Nuke trailing newline in hdstring */
    char * end = hdstring + strlen(hdstring) - 1;
    if (*end == '\n') {
        *end = '\0';
    }

#if HDSTRING_DUMP
    fprintf(stderr, "creating hdstring token \"%s\"\n", hdstring);
#endif

    /* sn_column() should be at the ; at the end of a heredoc */

    append_dqstring_token(hdstring,
            mode_start_line,
            mode_start_column,
            sn_line(),
            sn_column());

    /* A heredoc string is followed by a SEMICOLON */

    append_token(SEMICOLON, NULL,
            sn_line(),
            sn_column(),
            sn_line(),
            sn_column()+1);

    mode_buff.free(&mode_buff);
}


/* Helper method to enter HTML mode.
 * Should be called in place of BEGIN(HTML_MODE) */

void enter_html_mode() {
    if (token_dump_file) {
      LongStringInit(&mode_buff,0);
      mode_start_line = sn_line();
      mode_start_column = sn_column();
    }
    BEGIN(HTML_MODE);
}


void emit_html() {
    if (token_dump_file) {
      char * html_text = mode_buff.buf;

#if HTML_DUMP
      fprintf(stderr, "creating html token \"%s\"\n", html_text);
#endif

      append_token(HTML, html_text,
            mode_start_line,
            mode_start_column,
            sn_line(),
            sn_column());

      mode_buff.free(&mode_buff);
    }
}

/* Write out read xrefs for variables embedded in this
 * VDOUBLE_QUOTED_STRING token
 */

void emit_var_access_for_dqstring(Token* tok) {
  assert(tok->vars);
  for (tok = tok->vars; tok; tok = tok->next) {
      emit_var_access(tok, VAR_READ);
  }
}

/* This method is invoked when a var write operation is matched.
 * in the token stream.
 */

void emit_var_access(Token *tok, VarAccess acc) {
  char* varname = tok->strval;
  SearchEntry entry;
  int ref_to_symbol_type, ref_to_symbol_scope;
  int ref_from_scope_type;
  int line_start, line_end, column_start, column_end;

  line_start = tok->start_line;
  column_start = tok->start_column;
  line_end = tok->end_line;
  column_end = tok->end_column;

  /*
   * A var is global if not currently in a function,
   * if the variable is in the super global table,
   * or if in the global table.
   */

  entry.key = varname;
  entry.key_len = -1;

  if ((current_function == NULL) ||
      (super_global_var_table->search( &super_global_var_table, entry ) != NULL) ||
      (global_var_table &&
        (global_var_table->search( &global_var_table, entry ) != NULL))) {
    ref_to_symbol_type = SN_REF_TO_GLOB_VAR;
    ref_to_symbol_scope = SN_REF_SCOPE_GLOBAL;
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "global var \"%s\"\n", varname);
#endif
  } else {
    ref_to_symbol_type = SN_REF_TO_LOCAL_VAR;
    ref_to_symbol_scope = SN_REF_SCOPE_LOCAL;
#ifdef TOKEN_DEBUG
    fprintf(tokenout, "local var \"%s\"\n", varname);
#endif
  }

  /*
   * Pass SN_GLOBAL_NAMESPACE if the xref is outside of
   * a function, otherwise pass SN_FUNC_DEF.
   */   

  if (current_function == NULL) {
    ref_from_scope_type = SN_GLOBAL_NAMESPACE;
  } else {
    ref_from_scope_type = SN_FUNC_DEF;
  }

  if ((ref_to_symbol_type == SN_REF_TO_GLOB_VAR) ||
      ((ref_to_symbol_type == SN_REF_TO_LOCAL_VAR) &&
          ((int) sn_getopt(SN_OPT_LOCAL_VARS) != 0))) {

    if ((acc == VAR_READ) || (acc == VAR_READWRITE)) {
        result = sn_insert_xref(ref_to_symbol_type,
                 ref_from_scope_type,
                 ref_to_symbol_scope,
                 NULL,
                 current_function,
                 NULL,
                 NULL,
                 varname,
                 "UNDECLARED",
                 sn_current_file(),
                 line_start,
                 SN_REF_READ);

        assert(result == 0);
    }

    if ((acc == VAR_WRITE) || (acc == VAR_READWRITE)) {
        result = sn_insert_xref(ref_to_symbol_type,
                 ref_from_scope_type,
                 ref_to_symbol_scope,
                 NULL,
                 current_function,
                 NULL,
                 NULL,
                 varname,
                 "UNDECLARED",
                 sn_current_file(),
                 line_start,
                 SN_REF_WRITE);

        assert(result == 0);
    }
  }

  if (ref_to_symbol_type == SN_REF_TO_GLOB_VAR) {
      sn_highlight(SN_HIGH_VAR_GLOBAL,
          line_start, column_start,
          line_end, column_end);
  } else {
      sn_highlight(SN_HIGH_VAR_LOCAL,
          line_start, column_start,
          line_end, column_end);
  }
}

int get_num_tokens_matched(char *match) {
  char *x;
  int tokens;

  assert(*match != '\0'); /* match must have some text */
  assert(*match != ' '); /* match can't start with a space */
  for (x=match, tokens=0; *x; x++) {
    if (*x == ' ') {
        /* We just got to the end of a token */
        assert(*(x-1) != ' '); /* two spaces in a row is not allowed */
        tokens++;
    }
  }
  if (*(x-1) != ' ') {
    tokens++;
  }
  return tokens;
}

void
reset()
{
  assert(!current_function);
  sn_reset_line();
  sn_reset_column();
  sn_reset_encoding();
}

int
main(int argc, char *argv[])
{
  /* Init the super globals lookup table before starting
   * since it will never change during parsing.
   */

  int i;
  SearchEntry entry;
  char* super_globals[] = {
    "GLOBALS", "_SERVER", "_GET", "_POST", "_COOKIE",
    "_FILES", "_ENV", "_REQUEST", "_SESSION", NULL
  };
  super_global_var_table =
    SearchTableCreate(20, SEARCH_HASH_TABLE, FreeGlobalEntry);

  for (i=0; super_globals[i]; i++) {
    entry.key = super_globals[i];
    entry.key_len = -1;
    entry.flag = SEARCH_DUP_KEY; /* add copy of entry.key */
    super_global_var_table->add( &super_global_var_table, entry );
  }

  return sn_main(argc, argv, group, &yyin, yylex, reset);
}
